from abc import ABC
from typing import List, Tuple, Optional

import numpy as np

from sziszapangma.core.wer.distance_matrix_calculator import \
    DistanceCalculator
from sziszapangma.core.wer.step_type import StepType
from sziszapangma.core.wer.wer_processing_step import WerProcessingStep
from sziszapangma.core.wer.wer_span_question import Span
from sziszapangma.core.wer.wer_step import WerStep, StepWords
from sziszapangma.core.wer.word import Word


class WerCalculator(ABC):
    _distance_matrix_calculator: DistanceCalculator

    def __init__(self, distance_matrix_calculator: DistanceCalculator):
        self._distance_matrix_calculator = distance_matrix_calculator

    def convert_processing_steps_to_result(
        self,
        processing_steps: List[WerProcessingStep],
        reference_weights: Optional[List[float]] = None
    ) -> List[WerStep]:
        if reference_weights is None:
            return [
                WerStep(step.step_type, step.step_words, step.step_cost)
                for step in processing_steps
            ]
        else:
            indexes_per_steps = self._get_reference_indexes_per_steps(
                processing_steps)
            return [
                WerStep(
                    processing_steps[step_index].step_type,
                    processing_steps[step_index].step_words,
                    reference_weights[indexes_per_steps[step_index]] *
                    processing_steps[step_index].step_cost
                )
                for step_index in range(len(processing_steps))
            ]

    def get_distance_matrix_between_words(
        self,
        reference: List[Word],
        hypothesis: List[Word]
    ) -> np.ndarray:
        return self._distance_matrix_calculator.calculate_distance_matrix(
            reference, hypothesis)

    def extract_steps_path(
        self,
        steps_matrix: List[List[WerProcessingStep]]
    ) -> List[WerProcessingStep]:
        x = len(steps_matrix) - 1
        y = len(steps_matrix[0]) - 1
        to_return = []
        while not (x == 0 and y == 0):
            current_step = steps_matrix[x][y]
            to_return.append(current_step)
            if current_step.step_type == StepType.DELETION:
                x = x - 1
            elif current_step.step_type == StepType.INSERTION:
                y = y - 1
            else:  # creation and substitution
                y = y - 1
                x = x - 1
        return to_return[::-1]

    @staticmethod
    def _get_levenshtein_processing_step_cross(
        prev_cross_distance: float,
        step_words: StepWords,
        current_distance: float
    ) -> WerProcessingStep:
        return WerProcessingStep.levenshtein_correct(
            prev_cross_distance, step_words, 0) \
            if current_distance == 0 \
            else WerProcessingStep.levenshtein_substitution(
            prev_cross_distance, step_words, current_distance)

    def get_levenshtein_embedding_based(
        self,
        reference: List[Word],
        hypothesis: List[Word],
        distance_matrix: np.ndarray
    ) -> Tuple[np.ndarray, List[List[WerProcessingStep]]]:

        reference_len = len(reference)
        hypothesis_len = len(hypothesis)
        distance_arr, steps_arr = self._get_initialized_levenshtein_matrix(
            reference, hypothesis)

        for ref_index in range(reference_len):
            for hyp_index in range(hypothesis_len):
                step_words = StepWords(reference[ref_index],
                                       hypothesis[hyp_index])
                current_distance = distance_matrix[ref_index][hyp_index]
                prev_cross_distance = distance_arr[ref_index][hyp_index]

                cross_go_step = self._get_levenshtein_processing_step_cross(
                    prev_cross_distance, step_words, current_distance)
                insertion_step = WerProcessingStep.levenshtein_insertion(
                    distance_arr[ref_index + 1][hyp_index], step_words)
                deletion_step = WerProcessingStep.levenshtein_deletion(
                    distance_arr[ref_index][hyp_index + 1], step_words)

                best_step = min([cross_go_step, insertion_step, deletion_step],
                                key=lambda it: it.total_distance())

                distance_arr[ref_index + 1][hyp_index + 1] = \
                    best_step.total_distance()
                steps_arr[ref_index + 1][hyp_index + 1] = best_step

        return distance_arr, steps_arr

    @staticmethod
    def _get_initialized_levenshtein_matrix(
        reference: List[Word],
        hypothesis: List[Word]
    ) -> Tuple[np.ndarray, List[List[Optional[WerProcessingStep]]]]:

        # TODO: consider about remove distance_arr replaced by steps_arr
        reference_len = len(reference)
        hypothesis_len = len(hypothesis)
        distance_arr = np.zeros((reference_len + 1) * (hypothesis_len + 1)) \
            .reshape((reference_len + 1, hypothesis_len + 1))
        steps_arr = [
            [None for _ in range(hypothesis_len + 1)]
            for _ in range(reference_len + 1)
        ]

        # levenshtein initial
        for ref_index in range(reference_len + 1):
            distance_arr[ref_index][0] = ref_index
            step_words = StepWords(
                reference[ref_index - 1] if ref_index > 0 else None,
                None
            )
            steps_arr[ref_index][0] = WerProcessingStep.levenshtein_deletion(
                ref_index - 1, step_words)
        for hyp_index in range(hypothesis_len + 1):
            distance_arr[0][hyp_index] = hyp_index
            step_words = StepWords(
                None,
                hypothesis[hyp_index - 1] if hyp_index > 0 else None
            )
            steps_arr[0][hyp_index] = WerProcessingStep.levenshtein_insertion(
                hyp_index - 1, step_words)

        return distance_arr, steps_arr

    def _get_reference_indexes_per_steps(
        self,
        steps: List[WerProcessingStep]
    ) -> List[int]:
        counter = 0
        indexes = []
        for step in steps:
            indexes.append(counter)
            if step.step_type.contain_reference_word():
                counter = counter + 1
        return indexes

    def _calculate_wer(
        self,
        steps: List[WerStep],
    ) -> float:
        reference_len = sum([
            1 if step.step_type.contain_reference_word() else 0
            for step in steps
        ])
        return sum([step.step_cost for step in steps]) / reference_len

    def _calculate_steps_path(
        self,
        reference: List[Word],
        hypothesis: List[Word]
    ) -> List[WerProcessingStep]:
        distance_between_words = self.get_distance_matrix_between_words(
            reference, hypothesis)
        _, steps_matrix = self.get_levenshtein_embedding_based(
            reference, hypothesis, distance_between_words)
        return self.extract_steps_path(steps_matrix)

    def calculate_wer(
        self,
        reference: List[Word],
        hypothesis: List[Word]
    ) -> Tuple[float, List[WerStep]]:
        steps_path = self._calculate_steps_path(reference, hypothesis)
        steps = self.convert_processing_steps_to_result(steps_path)
        return self._calculate_wer(steps), steps

    def calculate_wer_for_spans(
        self,
        reference: List[Word],
        hypothesis: List[Word],
        spans: List[Span]
    ) -> List[float]:
        steps_path = self._calculate_steps_path(reference, hypothesis)
        reference_len = len(reference)
        return [
            self._calculate_wer(self.convert_processing_steps_to_result(
                processing_steps=steps_path,
                reference_weights=span.get_reference_weights_table(
                    reference_len)
            ))
            for span in spans
        ]

    def calculate_wer_weighted(
        self,
        reference: List[Word],
        hypothesis: List[Word],
        weights: List[float]
    ) -> Tuple[float, List[WerStep]]:
        steps_path = self._calculate_steps_path(reference, hypothesis)
        steps = self.convert_processing_steps_to_result(steps_path, weights)
        return self._calculate_wer(steps), steps
