from typing import List, Dict

from sziszapangma.core.transformer.cached_embedding_transformer import \
    CachedEmbeddingTransformer
from sziszapangma.core.transformer.embedding_transformer import \
    EmbeddingTransformer
from sziszapangma.core.wer.wer_embedding_calculator import \
    WerEmbeddingCalculator
from sziszapangma.core.wer.wer_soft_calculator import WerSoftCalculator
from sziszapangma.integration.repository.experiment_repository import \
    ExperimentRepository
from sziszapangma.integration.task.processing_task import ProcessingTask

_SOFT_WER = 'soft_wer'
_EMBEDDING_WER = 'embedding_wer'
_WORD = 'word'


class EmbeddingWerMetricsTask(ProcessingTask):
    _metrics_property_name: str
    _gold_transcript_property_name: str
    _embedding_transformer: CachedEmbeddingTransformer
    _wer_embedding_calculator: WerEmbeddingCalculator
    _wer_soft_calculator: WerSoftCalculator

    def __init__(
        self,
        task_name: str,
        gold_transcript_property_name: str,
        asr_property_name: str,
        metrics_property_name: str,
        require_update: bool,
        embedding_transformer: EmbeddingTransformer
    ):
        super().__init__(task_name, require_update)
        self._gold_transcript_property_name = gold_transcript_property_name
        self._asr_property_name = asr_property_name
        self._metrics_property_name = metrics_property_name
        self._embedding_transformer = \
            CachedEmbeddingTransformer(embedding_transformer)
        self._wer_embedding_calculator = \
            WerEmbeddingCalculator(self._embedding_transformer)
        self._wer_soft_calculator = \
            WerSoftCalculator(self._embedding_transformer)

    def skip_for_record(self, record_id: str,
                        experiment_repository: ExperimentRepository) -> bool:
        return experiment_repository \
            .get_property_for_key(record_id, self._metrics_property_name)

    def run_single_process(self, record_id: str,
                           experiment_repository: ExperimentRepository):
        gold_transcript = experiment_repository \
            .get_property_for_key(record_id,
                                  self._gold_transcript_property_name)
        asr_result = experiment_repository \
            .get_property_for_key(record_id, self._asr_property_name)
        if 'transcription' in asr_result:
            experiment_repository.update_property_for_key(
                record_id,
                self._metrics_property_name,
                self.calculate_metrics(
                    gold_transcript=gold_transcript,
                    asr_result=asr_result['transcription']
                )
            )
        self._embedding_transformer.clear()

    def calculate_metrics(
        self,
        gold_transcript: List[Dict[str, any]],
        asr_result: List[str]
    ) -> Dict[str, any]:
        """Calculate all metrics for data sample."""
        metrics = dict()
        reference = [it[_WORD] for it in gold_transcript]
        metrics[_SOFT_WER] = self._wer_soft_calculator.calculate_wer(
            reference, asr_result)[0]
        metrics[_EMBEDDING_WER] = self._wer_embedding_calculator.calculate_wer(
            reference, asr_result)[0]
        return metrics
